@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -57,10 +61,11 @@ static OFRunLoop *mainRunLoop = nil;
 #ifdef OF_HAVE_THREADS
 	OFMutex *_timersQueueMutex;
 #endif
-#if defined(OF_HAVE_SOCKETS)
+#ifdef OF_HAVE_SOCKETS
 	OFKernelEventObserver *_kernelEventObserver;
 	OFMutableDictionary *_readQueues, *_writeQueues;
-#elif defined(OF_HAVE_THREADS)
+#endif
+#ifdef OF_HAVE_THREADS
 	OFCondition *_condition;
 # ifdef OF_AMIGAOS
 	ULONG _execSignalMask;
@@ -91,7 +96,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFStreamAsyncReadBlock _block;
+	OFStreamReadHandler _handler;
 # endif
 	void *_buffer;
 	size_t _length;
@@ -102,18 +107,28 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFStreamAsyncReadBlock _block;
+	OFStreamReadHandler _handler;
 # endif
 	void *_buffer;
 	size_t _exactLength, _readLength;
 }
 @end
 
+@interface OFRunLoopReadStringQueueItem: OFRunLoopQueueItem
+{
+@public
+# ifdef OF_HAVE_BLOCKS
+	OFStreamStringReadHandler _handler;
+# endif
+	OFStringEncoding _encoding;
+}
+@end
+
 @interface OFRunLoopReadLineQueueItem: OFRunLoopQueueItem
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFStreamAsyncReadLineBlock _block;
+	OFStreamStringReadHandler _handler;
 # endif
 	OFStringEncoding _encoding;
 }
@@ -123,7 +138,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFStreamAsyncWriteDataBlock _block;
+	OFStreamDataWrittenHandler _handler;
 # endif
 	OFData *_data;
 	size_t _writtenLength;
@@ -134,7 +149,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFStreamAsyncWriteStringBlock _block;
+	OFStreamStringWrittenHandler _handler;
 # endif
 	OFString *_string;
 	OFStringEncoding _encoding;
@@ -151,7 +166,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	id _block;
+	id _handler;
 # endif
 }
 @end
@@ -160,7 +175,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFDatagramSocketAsyncReceiveBlock _block;
+	OFDatagramSocketPacketReceivedHandler _handler;
 # endif
 	void *_buffer;
 	size_t _length;
@@ -171,7 +186,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFDatagramSocketAsyncSendDataBlock _block;
+	OFDatagramSocketDataSentHandler _handler;
 # endif
 	OFData *_data;
 	OFSocketAddress _receiver;
@@ -182,7 +197,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFSequencedPacketSocketAsyncReceiveBlock _block;
+	OFSequencedPacketSocketPacketReceivedHandler _handler;
 # endif
 	void *_buffer;
 	size_t _length;
@@ -193,11 +208,35 @@ static OFRunLoop *mainRunLoop = nil;
 {
 @public
 # ifdef OF_HAVE_BLOCKS
-	OFSequencedPacketSocketAsyncSendDataBlock _block;
+	OFSequencedPacketSocketDataSentHandler _handler;
 # endif
 	OFData *_data;
 }
 @end
+
+# ifdef OF_HAVE_SCTP
+@interface OFRunLoopSCTPReceiveQueueItem: OFRunLoopQueueItem
+{
+@public
+#  ifdef OF_HAVE_BLOCKS
+	OFSCTPSocketMessageReceivedHandler _handler;
+#  endif
+	void *_buffer;
+	size_t _length;
+}
+@end
+
+@interface OFRunLoopSCTPSendQueueItem: OFRunLoopQueueItem
+{
+@public
+#  ifdef OF_HAVE_BLOCKS
+	OFSCTPSocketDataSentHandler _handler;
+#  endif
+	OFData *_data;
+	OFSCTPMessageInfo _info;
+}
+@end
+# endif
 #endif
 
 @implementation OFRunLoopState
@@ -211,13 +250,11 @@ static OFRunLoop *mainRunLoop = nil;
 		_timersQueueMutex = [[OFMutex alloc] init];
 #endif
 
-#if defined(OF_HAVE_SOCKETS)
-		_kernelEventObserver = [[OFKernelEventObserver alloc] init];
-		_kernelEventObserver.delegate = self;
-
+#ifdef OF_HAVE_SOCKETS
 		_readQueues = [[OFMutableDictionary alloc] init];
 		_writeQueues = [[OFMutableDictionary alloc] init];
-#elif defined(OF_HAVE_THREADS)
+#endif
+#if defined(OF_HAVE_THREADS)
 		_condition = [[OFCondition alloc] init];
 #endif
 #ifdef OF_AMIGAOS
@@ -244,11 +281,12 @@ static OFRunLoop *mainRunLoop = nil;
 #ifdef OF_HAVE_THREADS
 	[_timersQueueMutex release];
 #endif
-#if defined(OF_HAVE_SOCKETS)
+#ifdef OF_HAVE_SOCKETS
 	[_kernelEventObserver release];
 	[_readQueues release];
 	[_writeQueues release];
-#elif defined(OF_HAVE_THREADS)
+#endif
+#ifdef OF_HAVE_THREADS
 	[_condition release];
 #endif
 #ifdef OF_AMIGAOS
@@ -433,8 +471,8 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL)
-		return _block(length, exception);
+	if (_handler != NULL)
+		return _handler(object, _buffer, length, exception);
 	else {
 # endif
 		if (![_delegate respondsToSelector:
@@ -453,7 +491,7 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
 
 	[super dealloc];
 }
@@ -481,8 +519,8 @@ static OFRunLoop *mainRunLoop = nil;
 		return true;
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
-		if (!_block(_readLength, exception))
+	if (_handler != NULL) {
+		if (!_handler(object, _buffer, _readLength, exception))
 			return false;
 
 		_readLength = 0;
@@ -509,7 +547,50 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
+
+	[super dealloc];
+}
+# endif
+@end
+
+@implementation OFRunLoopReadStringQueueItem
+- (bool)handleObject: (id)object
+{
+	OFString *string;
+	id exception = nil;
+
+	@try {
+		string = [object tryReadStringWithEncoding: _encoding];
+	} @catch (id e) {
+		string = nil;
+		exception = e;
+	}
+
+	if (string == nil && ![object isAtEndOfStream] && exception == nil)
+		return true;
+
+# ifdef OF_HAVE_BLOCKS
+	if (_handler != NULL)
+		return _handler(object, string, exception);
+	else {
+# endif
+		if (![_delegate respondsToSelector:
+		    @selector(stream:didReadString:exception:)])
+			return false;
+
+		return [_delegate stream: object
+			   didReadString: string
+			       exception: exception];
+# ifdef OF_HAVE_BLOCKS
+	}
+# endif
+}
+
+# ifdef OF_HAVE_BLOCKS
+- (void)dealloc
+{
+	[_handler release];
 
 	[super dealloc];
 }
@@ -533,8 +614,8 @@ static OFRunLoop *mainRunLoop = nil;
 		return true;
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL)
-		return _block(line, exception);
+	if (_handler != NULL)
+		return _handler(object, line, exception);
 	else {
 # endif
 		if (![_delegate respondsToSelector:
@@ -552,7 +633,7 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
 
 	[super dealloc];
 }
@@ -588,8 +669,8 @@ static OFRunLoop *mainRunLoop = nil;
 		return true;
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
-		newData = _block(_writtenLength, exception);
+	if (_handler != NULL) {
+		newData = _handler(object, _data, _writtenLength, exception);
 
 		if (newData == nil)
 			return false;
@@ -627,10 +708,10 @@ static OFRunLoop *mainRunLoop = nil;
 
 - (void)dealloc
 {
-	[_data release];
 # ifdef OF_HAVE_BLOCKS
-	[_block release];
+	[_handler release];
 # endif
+	[_data release];
 
 	[super dealloc];
 }
@@ -665,8 +746,9 @@ static OFRunLoop *mainRunLoop = nil;
 		return true;
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
-		newString = _block(_writtenLength, exception);
+	if (_handler != NULL) {
+		newString = _handler(object, _string, _encoding, _writtenLength,
+		    exception);
 
 		if (newString == nil)
 			return false;
@@ -707,7 +789,7 @@ static OFRunLoop *mainRunLoop = nil;
 {
 	[_string release];
 # ifdef OF_HAVE_BLOCKS
-	[_block release];
+	[_handler release];
 # endif
 
 	[super dealloc];
@@ -765,14 +847,14 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
+	if (_handler != NULL) {
 		if ([object isKindOfClass: [OFStreamSocket class]])
-			return ((OFStreamSocketAsyncAcceptBlock)_block)(
-			    acceptedSocket, exception);
+			return ((OFStreamSocketAcceptedHandler)
+			    _handler)(object, acceptedSocket, exception);
 		else if ([object isKindOfClass:
 		    [OFSequencedPacketSocket class]])
-			return ((OFSequencedPacketSocketAsyncAcceptBlock)
-			    _block)(acceptedSocket, exception);
+			return ((OFSequencedPacketSocketAcceptedHandler)
+			    _handler)(object, acceptedSocket, exception);
 		else
 			OFEnsure(0);
 	} else {
@@ -792,7 +874,7 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
 
 	[super dealloc];
 }
@@ -816,8 +898,8 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL)
-		return _block(length, &address, exception);
+	if (_handler != NULL)
+		return _handler(object, _buffer, length, &address, exception);
 	else {
 # endif
 		if (![_delegate respondsToSelector: @selector(
@@ -837,7 +919,7 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
 
 	[super dealloc];
 }
@@ -859,8 +941,8 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
-		newData = _block(exception);
+	if (_handler != NULL) {
+		newData = _handler(object, _data, &_receiver, exception);
 
 		if (newData == nil)
 			return false;
@@ -896,10 +978,10 @@ static OFRunLoop *mainRunLoop = nil;
 
 - (void)dealloc
 {
-	[_data release];
 # ifdef OF_HAVE_BLOCKS
-	[_block release];
+	[_handler release];
 # endif
+	[_data release];
 
 	[super dealloc];
 }
@@ -919,8 +1001,8 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL)
-		return _block(length, exception);
+	if (_handler != NULL)
+		return _handler(object, _buffer, length, exception);
 	else {
 # endif
 		if (![_delegate respondsToSelector: @selector(
@@ -939,7 +1021,7 @@ static OFRunLoop *mainRunLoop = nil;
 # ifdef OF_HAVE_BLOCKS
 - (void)dealloc
 {
-	[_block release];
+	[_handler release];
 
 	[super dealloc];
 }
@@ -960,8 +1042,8 @@ static OFRunLoop *mainRunLoop = nil;
 	}
 
 # ifdef OF_HAVE_BLOCKS
-	if (_block != NULL) {
-		newData = _block(exception);
+	if (_handler != NULL) {
+		newData = _handler(object, _data, exception);
 
 		if (newData == nil)
 			return false;
@@ -996,14 +1078,123 @@ static OFRunLoop *mainRunLoop = nil;
 
 - (void)dealloc
 {
-	[_data release];
 # ifdef OF_HAVE_BLOCKS
-	[_block release];
+	[_handler release];
 # endif
+	[_data release];
 
 	[super dealloc];
 }
 @end
+
+# ifdef OF_HAVE_SCTP
+@implementation OFRunLoopSCTPReceiveQueueItem
+- (bool)handleObject: (id)object
+{
+	size_t length;
+	OFSCTPMessageInfo info;
+	id exception = nil;
+
+	@try {
+		length = [object receiveIntoBuffer: _buffer
+					    length: _length
+					      info: &info];
+	} @catch (id e) {
+		length = 0;
+		exception = e;
+	}
+
+#  ifdef OF_HAVE_BLOCKS
+	if (_handler != NULL)
+		return _handler(object, _buffer, length, info, exception);
+	else {
+#  endif
+		if (![_delegate respondsToSelector: @selector(
+		    socket:didReceiveIntoBuffer:length:info:exception:)])
+			return false;
+
+		return [_delegate socket: object
+		    didReceiveIntoBuffer: _buffer
+				  length: length
+				    info: info
+			       exception: exception];
+#  ifdef OF_HAVE_BLOCKS
+	}
+#  endif
+}
+
+# ifdef OF_HAVE_BLOCKS
+- (void)dealloc
+{
+	[_handler release];
+
+	[super dealloc];
+}
+# endif
+@end
+
+@implementation OFRunLoopSCTPSendQueueItem
+- (bool)handleObject: (id)object
+{
+	id exception = nil;
+	OFData *newData, *oldData;
+
+	@try {
+		[object sendBuffer: _data.items
+			    length: _data.count * _data.itemSize
+			      info: _info];
+	} @catch (id e) {
+		exception = e;
+	}
+
+#  ifdef OF_HAVE_BLOCKS
+	if (_handler != NULL) {
+		newData = _handler(object, _data, _info, exception);
+
+		if (newData == nil)
+			return false;
+
+		oldData = _data;
+		_data = [newData copy];
+		[oldData release];
+
+		return true;
+	} else {
+#  endif
+		if (![_delegate respondsToSelector: @selector(
+		    socket:didSendData:info:exception:)])
+			return false;
+
+		newData = [_delegate socket: object
+				didSendData: _data
+				       info: _info
+				  exception: exception];
+
+		if (newData == nil)
+			return false;
+
+		oldData = _data;
+		_data = [newData copy];
+		[oldData release];
+
+		return true;
+#  ifdef OF_HAVE_BLOCKS
+	}
+#  endif
+}
+
+- (void)dealloc
+{
+# ifdef OF_HAVE_BLOCKS
+	[_handler release];
+# endif
+	[_data release];
+	[_info release];
+
+	[super dealloc];
+}
+@end
+# endif
 #endif
 
 @implementation OFRunLoop
@@ -1029,7 +1220,8 @@ static OFRunLoop *mainRunLoop = nil;
 }
 
 static OFRunLoopState *
-stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
+stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create,
+    bool createObserver)
 {
 	OFRunLoopState *state;
 
@@ -1047,6 +1239,14 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 				[state release];
 			}
 		}
+
+#ifdef OF_HAVE_SOCKETS
+		if (createObserver && state->_kernelEventObserver == nil) {
+			state->_kernelEventObserver =
+			    [[OFKernelEventObserver alloc] init];
+			state->_kernelEventObserver.delegate = state;
+		}
+#endif
 #ifdef OF_HAVE_THREADS
 	} @finally {
 		[self->_statesMutex unlock];
@@ -1057,43 +1257,43 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 }
 
 #ifdef OF_HAVE_SOCKETS
-# define NEW_READ(type, object, mode)					\
-	void *pool = objc_autoreleasePoolPush();			\
-	OFRunLoop *runLoop = [self currentRunLoop];			\
-	OFRunLoopState *state = stateForMode(runLoop, mode, true);	\
-	OFList *queue = [state->_readQueues objectForKey: object];	\
-	type *queueItem;						\
-									\
-	if (queue == nil) {						\
-		queue = [OFList list];					\
-		[state->_readQueues setObject: queue forKey: object];	\
-	}								\
-									\
-	if (queue.count == 0)						\
-		[state->_kernelEventObserver				\
-		    addObjectForReading: object];			\
-									\
+# define NEW_READ(type, object, mode)					 \
+	void *pool = objc_autoreleasePoolPush();			 \
+	OFRunLoop *runLoop = [self currentRunLoop];			 \
+	OFRunLoopState *state = stateForMode(runLoop, mode, true, true); \
+	OFList *queue = [state->_readQueues objectForKey: object];	 \
+	type *queueItem;						 \
+									 \
+	if (queue == nil) {						 \
+		queue = [OFList list];					 \
+		[state->_readQueues setObject: queue forKey: object];	 \
+	}								 \
+									 \
+	if (queue.count == 0)						 \
+		[state->_kernelEventObserver				 \
+		    addObjectForReading: object];			 \
+									 \
 	queueItem = [[[type alloc] init] autorelease];
-# define NEW_WRITE(type, object, mode)					\
-	void *pool = objc_autoreleasePoolPush();			\
-	OFRunLoop *runLoop = [self currentRunLoop];			\
-	OFRunLoopState *state = stateForMode(runLoop, mode, true);	\
-	OFList *queue = [state->_writeQueues objectForKey: object];	\
-	type *queueItem;						\
-									\
-	if (queue == nil) {						\
-		queue = [OFList list];					\
-		[state->_writeQueues setObject: queue forKey: object];	\
-	}								\
-									\
-	if (queue.count == 0)						\
-		[state->_kernelEventObserver				\
-		    addObjectForWriting: object];			\
-									\
+# define NEW_WRITE(type, object, mode)					 \
+	void *pool = objc_autoreleasePoolPush();			 \
+	OFRunLoop *runLoop = [self currentRunLoop];			 \
+	OFRunLoopState *state = stateForMode(runLoop, mode, true, true); \
+	OFList *queue = [state->_writeQueues objectForKey: object];	 \
+	type *queueItem;						 \
+									 \
+	if (queue == nil) {						 \
+		queue = [OFList list];					 \
+		[state->_writeQueues setObject: queue forKey: object];	 \
+	}								 \
+									 \
+	if (queue.count == 0)						 \
+		[state->_kernelEventObserver				 \
+		    addObjectForWriting: object];			 \
+									 \
 	queueItem = [[[type alloc] init] autorelease];
-#define QUEUE_ITEM							\
-	[queue appendObject: queueItem];				\
-									\
+#define QUEUE_ITEM							 \
+	[queue appendObject: queueItem];				 \
+									 \
 	objc_autoreleasePoolPop(pool);
 
 + (void)of_addAsyncReadForStream: (OFStream <OFReadyForReadingObserving> *)
@@ -1102,7 +1302,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 			  length: (size_t)length
 			    mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-			   block: (OFStreamAsyncReadBlock)block
+			 handler: (OFStreamReadHandler)handler
 # endif
 			delegate: (id <OFStreamDelegate>)delegate
 {
@@ -1110,7 +1310,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_buffer = buffer;
 	queueItem->_length = length;
@@ -1124,7 +1324,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 		     exactLength: (size_t)exactLength
 			    mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-			   block: (OFStreamAsyncReadBlock)block
+			 handler: (OFStreamReadHandler)handler
 # endif
 			delegate: (id <OFStreamDelegate>)delegate
 {
@@ -1132,7 +1332,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_buffer = buffer;
 	queueItem->_exactLength = exactLength;
@@ -1140,12 +1340,32 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 	QUEUE_ITEM
 }
 
++ (void)of_addAsyncReadStringForStream: (OFStream <OFReadyForReadingObserving
+					    > *)stream
+			      encoding: (OFStringEncoding)encoding
+				  mode: (OFRunLoopMode)mode
+# ifdef OF_HAVE_BLOCKS
+			       handler: (OFStreamStringReadHandler)handler
+# endif
+			      delegate: (id <OFStreamDelegate>)delegate
+{
+	NEW_READ(OFRunLoopReadStringQueueItem, stream, mode)
+
+	queueItem->_delegate = [delegate retain];
+# ifdef OF_HAVE_BLOCKS
+	queueItem->_handler = [handler copy];
+# endif
+	queueItem->_encoding = encoding;
+
+	QUEUE_ITEM
+}
+
 + (void)of_addAsyncReadLineForStream: (OFStream <OFReadyForReadingObserving> *)
 					  stream
 			    encoding: (OFStringEncoding)encoding
 				mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-			       block: (OFStreamAsyncReadLineBlock)block
+			     handler: (OFStreamStringReadHandler)handler
 # endif
 			    delegate: (id <OFStreamDelegate>)delegate
 {
@@ -1153,7 +1373,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_encoding = encoding;
 
@@ -1165,7 +1385,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 			     data: (OFData *)data
 			     mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-			    block: (OFStreamAsyncWriteDataBlock)block
+			  handler: (OFStreamDataWrittenHandler)handler
 # endif
 			 delegate: (id <OFStreamDelegate>)delegate
 {
@@ -1173,7 +1393,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_data = [data copy];
 
@@ -1186,7 +1406,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 			 encoding: (OFStringEncoding)encoding
 			     mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-			    block: (OFStreamAsyncWriteStringBlock)block
+			  handler: (OFStreamStringWrittenHandler)handler
 # endif
 			 delegate: (id <OFStreamDelegate>)delegate
 {
@@ -1194,7 +1414,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_string = [string copy];
 	queueItem->_encoding = encoding;
@@ -1217,14 +1437,14 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 + (void)of_addAsyncAcceptForSocket: (id)sock
 			      mode: (OFRunLoopMode)mode
-			     block: (id)block
+			   handler: (id)handler
 			  delegate: (id)delegate
 {
 	NEW_READ(OFRunLoopAcceptQueueItem, sock, mode)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 
 	QUEUE_ITEM
@@ -1235,7 +1455,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
     length: (size_t)length
       mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-     block: (OFDatagramSocketAsyncReceiveBlock)block
+   handler: (OFDatagramSocketPacketReceivedHandler)handler
 # endif
   delegate: (id <OFDatagramSocketDelegate>)delegate
 {
@@ -1243,7 +1463,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_buffer = buffer;
 	queueItem->_length = length;
@@ -1256,7 +1476,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
   receiver: (const OFSocketAddress *)receiver
       mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-     block: (OFDatagramSocketAsyncSendDataBlock)block
+   handler: (OFDatagramSocketDataSentHandler)handler
 # endif
   delegate: (id <OFDatagramSocketDelegate>)delegate
 {
@@ -1264,7 +1484,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_data = [data copy];
 	queueItem->_receiver = *receiver;
@@ -1278,7 +1498,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
     length: (size_t)length
       mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-     block: (OFSequencedPacketSocketAsyncReceiveBlock)block
+   handler: (OFSequencedPacketSocketPacketReceivedHandler)handler
 # endif
   delegate: (id <OFSequencedPacketSocketDelegate>)delegate
 {
@@ -1286,7 +1506,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_buffer = buffer;
 	queueItem->_length = length;
@@ -1298,7 +1518,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
       data: (OFData *)data
       mode: (OFRunLoopMode)mode
 # ifdef OF_HAVE_BLOCKS
-     block: (OFSequencedPacketSocketAsyncSendDataBlock)block
+   handler: (OFSequencedPacketSocketDataSentHandler)handler
 # endif
   delegate: (id <OFSequencedPacketSocketDelegate>)delegate
 {
@@ -1306,12 +1526,57 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	queueItem->_delegate = [delegate retain];
 # ifdef OF_HAVE_BLOCKS
-	queueItem->_block = [block copy];
+	queueItem->_handler = [handler copy];
 # endif
 	queueItem->_data = [data copy];
 
 	QUEUE_ITEM
 }
+
+# ifdef OF_HAVE_SCTP
++ (void)
+    of_addAsyncReceiveForSCTPSocket: (OFSCTPSocket *)sock
+			     buffer: (void *)buffer
+			     length: (size_t)length
+			       mode: (OFRunLoopMode)mode
+#  ifdef OF_HAVE_BLOCKS
+			    handler: (OFSCTPSocketMessageReceivedHandler)handler
+#  endif
+			   delegate: (id <OFSCTPSocketDelegate>)delegate
+{
+	NEW_READ(OFRunLoopSCTPReceiveQueueItem, sock, mode)
+
+	queueItem->_delegate = [delegate retain];
+#  ifdef OF_HAVE_BLOCKS
+	queueItem->_handler = [handler copy];
+#  endif
+	queueItem->_buffer = buffer;
+	queueItem->_length = length;
+
+	QUEUE_ITEM
+}
+
++ (void)of_addAsyncSendForSCTPSocket: (OFSCTPSocket *)sock
+				data: (OFData *)data
+				info: (OFSCTPMessageInfo)info
+				mode: (OFRunLoopMode)mode
+# ifdef OF_HAVE_BLOCKS
+			     handler: (OFSCTPSocketDataSentHandler)handler
+# endif
+			    delegate: (id <OFSCTPSocketDelegate>)delegate
+{
+	NEW_WRITE(OFRunLoopSCTPSendQueueItem, sock, mode)
+
+	queueItem->_delegate = [delegate retain];
+# ifdef OF_HAVE_BLOCKS
+	queueItem->_handler = [handler copy];
+# endif
+	queueItem->_data = [data copy];
+	queueItem->_info = [info copy];
+
+	QUEUE_ITEM
+}
+# endif
 # undef NEW_READ
 # undef NEW_WRITE
 # undef QUEUE_ITEM
@@ -1320,7 +1585,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 {
 	void *pool = objc_autoreleasePoolPush();
 	OFRunLoop *runLoop = [self currentRunLoop];
-	OFRunLoopState *state = stateForMode(runLoop, mode, false);
+	OFRunLoopState *state = stateForMode(runLoop, mode, false, false);
 	OFList *queue;
 
 	if (state == nil)
@@ -1400,7 +1665,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 - (void)addTimer: (OFTimer *)timer forMode: (OFRunLoopMode)mode
 {
-	OFRunLoopState *state = stateForMode(self, mode, true);
+	OFRunLoopState *state = stateForMode(self, mode, true, false);
 
 #ifdef OF_HAVE_THREADS
 	[state->_timersQueueMutex lock];
@@ -1415,16 +1680,17 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 	[timer of_setInRunLoop: self mode: mode];
 
-#if defined(OF_HAVE_SOCKETS)
+#ifdef OF_HAVE_SOCKETS
 	[state->_kernelEventObserver cancel];
-#elif defined(OF_HAVE_THREADS)
+#endif
+#ifdef OF_HAVE_THREADS
 	[state->_condition signal];
 #endif
 }
 
 - (void)of_removeTimer: (OFTimer *)timer forMode: (OFRunLoopMode)mode
 {
-	OFRunLoopState *state = stateForMode(self, mode, false);
+	OFRunLoopState *state = stateForMode(self, mode, false, false);
 
 	/* {} required to avoid -Wmisleading-indentation false positive. */
 	if (state == nil) {
@@ -1463,7 +1729,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 	       target: (id)target
 	     selector: (SEL)selector
 {
-	OFRunLoopState *state = stateForMode(self, mode, true);
+	OFRunLoopState *state = stateForMode(self, mode, true, false);
 
 # ifdef OF_HAVE_THREADS
 	[state->_execSignalsMutex lock];
@@ -1475,7 +1741,8 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 # ifdef OF_HAVE_SOCKETS
 		state->_kernelEventObserver.execSignalMask |= (1ul << signal);
-# elif defined(OF_HAVE_THREADS)
+# endif
+# ifdef OF_HAVE_THREADS
 		state->_execSignalMask |= (1ul << signal);
 # endif
 # ifdef OF_HAVE_THREADS
@@ -1484,9 +1751,10 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 	}
 # endif
 
-# if defined(OF_HAVE_SOCKETS)
+# ifdef OF_HAVE_SOCKETS
 	[state->_kernelEventObserver cancel];
-# elif defined(OF_HAVE_THREADS)
+# endif
+# ifdef OF_HAVE_THREADS
 	[state->_condition signal];
 # endif
 }
@@ -1506,7 +1774,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 		  target: (id)target
 		selector: (SEL)selector
 {
-	OFRunLoopState *state = stateForMode(self, mode, false);
+	OFRunLoopState *state = stateForMode(self, mode, false, false);
 
 	if (state == nil)
 		return;
@@ -1538,7 +1806,8 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 # ifdef OF_HAVE_SOCKETS
 		state->_kernelEventObserver.execSignalMask = newMask;
-# elif defined(OF_HAVE_THREADS)
+# endif
+# ifdef OF_HAVE_THREADS
 		state->_execSignalMask = newMask;
 # endif
 # ifdef OF_HAVE_THREADS
@@ -1547,9 +1816,10 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 	}
 # endif
 
-# if defined(OF_HAVE_SOCKETS)
+# ifdef OF_HAVE_SOCKETS
 	[state->_kernelEventObserver cancel];
-# elif defined(OF_HAVE_THREADS)
+# endif
+# ifdef OF_HAVE_THREADS
 	[state->_condition signal];
 # endif
 }
@@ -1573,7 +1843,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 {
 	void *pool = objc_autoreleasePoolPush();
 	OFRunLoopMode previousMode = _currentMode;
-	OFRunLoopState *state = stateForMode(self, mode, false);
+	OFRunLoopState *state = stateForMode(self, mode, false, false);
 
 	if (state == nil)
 		return;
@@ -1581,7 +1851,7 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 	_currentMode = mode;
 	@try {
 		OFDate *nextTimer;
-#if defined(OF_AMIGAOS) && !defined(OF_HAVE_SOCKETS) && defined(OF_HAVE_THREADS)
+#if defined(OF_AMIGAOS) && defined(OF_HAVE_THREADS)
 		ULONG signalMask;
 #endif
 
@@ -1644,31 +1914,42 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 				timeout = [nextTimer earlierDate: deadline]
 				    .timeIntervalSinceNow;
 
-			if (timeout < 0)
+			if (timeout < 0) {
 				timeout = 0;
-
-#if defined(OF_HAVE_SOCKETS)
-			@try {
-				[state->_kernelEventObserver
-				    observeForTimeInterval: timeout];
-			} @catch (OFObserveKernelEventsFailedException *e) {
-				if (e.errNo != EINTR)
-					@throw e;
 			}
-#elif defined(OF_HAVE_THREADS)
-			[state->_condition lock];
+
+#ifdef OF_HAVE_SOCKETS
+			if (state->_kernelEventObserver != nil) {
+				@try {
+					[state->_kernelEventObserver
+					    observeForTimeInterval: timeout];
+				} @catch (OFObserveKernelEventsFailedException
+				    *e) {
+					if (e.errNo != EINTR)
+						@throw e;
+				}
+			} else {
+#endif
+#ifdef OF_HAVE_THREADS
+				[state->_condition lock];
 # ifdef OF_AMIGAOS
-			signalMask = state->_execSignalMask;
-			[state->_condition waitForTimeInterval: timeout
-						  orExecSignal: &signalMask];
-			if (signalMask != 0)
-				[state execSignalWasReceived: signalMask];
+				signalMask = state->_execSignalMask;
+				[state->_condition
+				    waitForTimeInterval: timeout
+					   orExecSignal: &signalMask];
+				if (signalMask != 0)
+					[state
+					    execSignalWasReceived: signalMask];
 # else
-			[state->_condition waitForTimeInterval: timeout];
+				[state->_condition
+				    waitForTimeInterval: timeout];
 # endif
-			[state->_condition unlock];
+				[state->_condition unlock];
 #else
-			[OFThread sleepForTimeInterval: timeout];
+				[OFThread sleepForTimeInterval: timeout];
+#endif
+#ifdef OF_HAVE_SOCKETS
+			}
 #endif
 		} else {
 			/*
@@ -1676,27 +1957,35 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 			 * until we get an event. If a timer is added by
 			 * another thread, it cancels the observe.
 			 */
-#if defined(OF_HAVE_SOCKETS)
-			@try {
-				[state->_kernelEventObserver observe];
-			} @catch (OFObserveKernelEventsFailedException *e) {
-				if (e.errNo != EINTR)
-					@throw e;
-			}
-#elif defined(OF_HAVE_THREADS)
-			[state->_condition lock];
+#ifdef OF_HAVE_SOCKETS
+			if (state->_kernelEventObserver != nil) {
+				@try {
+					[state->_kernelEventObserver observe];
+				} @catch (OFObserveKernelEventsFailedException
+				    *e) {
+					if (e.errNo != EINTR)
+						@throw e;
+				}
+			} else {
+#endif
+#ifdef OF_HAVE_THREADS
+				[state->_condition lock];
 # ifdef OF_AMIGAOS
-			signalMask = state->_execSignalMask;
-			[state->_condition
-			    waitForConditionOrExecSignal: &signalMask];
-			if (signalMask != 0)
-				[state execSignalWasReceived: signalMask];
+				signalMask = state->_execSignalMask;
+				[state->_condition
+				    waitForConditionOrExecSignal: &signalMask];
+				if (signalMask != 0)
+					[state
+					    execSignalWasReceived: signalMask];
 # else
-			[state->_condition wait];
+				[state->_condition wait];
 # endif
-			[state->_condition unlock];
+				[state->_condition unlock];
 #else
-			[OFThread sleepForTimeInterval: 86400];
+				[OFThread sleepForTimeInterval: 86400];
+#endif
+#ifdef OF_HAVE_SOCKETS
+			}
 #endif
 		}
 
@@ -1708,16 +1997,18 @@ stateForMode(OFRunLoop *self, OFRunLoopMode mode, bool create)
 
 - (void)stop
 {
-	OFRunLoopState *state = stateForMode(self, OFDefaultRunLoopMode, false);
+	OFRunLoopState *state = stateForMode(self, OFDefaultRunLoopMode,
+	    false, false);
 
 	_stop = true;
 
 	if (state == nil)
 		return;
 
-#if defined(OF_HAVE_SOCKETS)
+#ifdef OF_HAVE_SOCKETS
 	[state->_kernelEventObserver cancel];
-#elif defined(OF_HAVE_THREADS)
+#endif
+#ifdef OF_HAVE_THREADS
 	[state->_condition signal];
 #endif
 }
